//conv.cpp by Denis L. Seleznev / Hidden Codes 2k4, mailto:hcodes@narod.ru

#include "Autocode.h"
#include "maps.c"
//#include "Cvtutf7.c"

#define MAX_RUS_SYMBOLS		4096
#define MAP_SIZE			256

#define MAX_UCHAR			256
#define MAX_LATCHAR		128
#define MAX_USHORT		65536

#define CC_1251			"1251"
#define CC_866			"866"
#define CC_DM			"DM"
#define CC_KOI8R			"KOI8R"
#define CC_GOST			"GOST"
#define CC_MAC			"MAC"
#define CC_UNICODE			"UNICODE"
#define CC_TRANSLIT		"TRANSLIT"
#define CC_TRANSLIT_USA		"TRANSLIT_USA"
#define CC_TRANSLIT_BGN		"TRANSLIT_BGN"
#define CC_NONE			"NONE"

// for future use
#define CC_UTF32		"UTF32"
#define CC_UTF16		"UTF16"
#define CC_UTF8		"UTF8"
#define CC_UTF7		"UTF7"

char* pMaps[]={
		CC_1251, CC_866, CC_DM, CC_KOI8R, CC_GOST, CC_MAC,
		CC_UNICODE, CC_TRANSLIT, CC_TRANSLIT_USA, CC_TRANSLIT_BGN,
		CC_NONE,
		/* CC_UTF32, CC_UTF_16, CC_UTF8,  CC_UTF7,*/
		""};

#define MAPS_N			(sizeof(pMaps)/sizeof(char*)-1)
#define MAPS_AUTO_N		6
#define MAPS_FOR_MESSAGEBOX_N		7


enum {CN_1251, CN_866, CN_DM, CN_KOI8R, CN_GOST, CN_MAC,
	CN_UNICODE, CN_TRANSLIT, CN_TRANSLIT_USA, CN_TRANSLIT_BGN,
	CN_NONE,
	CN_UTF32, CN_UTF_16, CN_UTF8, CN_UTF7};

char SizeMap[MAPS_N]=
	{
	1, // 1251
	1,	// 866
	1, // DM
	1, //KOI8
	1, //GOST
	1, //MAC
	2, //UNICODE
	3, //TRANSLIT
	4, //TRANSLIT_USA
	4, //TRANSLIT_BGN
	0};//NONE
			/* 	4, UTF32
				X, UTF16
				X, UTF8
				X, UTF7
				*/

unsigned long ConvertWithConstMap(char* szIn, char* szOut, unsigned char* ucText, unsigned long  ulLen, unsigned char* ucTextBuf, unsigned long ulLenBuf, char* szErr);
BOOL ConvertWithMapFile(char* szMap, unsigned char* ucText, unsigned long ulLen, unsigned char* ucTextBuf, unsigned long ulLenBuf, char* szErr);
BOOL DefineMap(unsigned char* ucText, unsigned long ulLen, char* szMap);
BOOL IsSupportMap(char* szMap);
unsigned long C1251toUNICODE(unsigned char* ucText, unsigned long ulLen, unsigned char* ucTextBuf, unsigned long ulLenBuf);
unsigned long UNICODEtoC1251(unsigned char* ucText, unsigned long ulLen, unsigned char* ucTextBuf, unsigned long ulLenBuf);
unsigned long CalcBufMap(char* szIn, char* szOut, unsigned char* ucText, unsigned ulLen);
unsigned long C1251toTranslit(unsigned char* ucText, unsigned long ulLen, unsigned char* ucTextBuf, unsigned long ulLenBuf, int iMode);

//---------------------------------------------------------------------------
BOOL ConvertWithMapFile(char* szMap, unsigned char* ucText, unsigned long ulLen, unsigned  char* ucTextBuf, unsigned long ulLenBuf, char* szErr)
{
	FILE *hCmFile=NULL;

	unsigned  char* ucTextBufEnd=ucTextBuf+ulLenBuf;

	char chBufCm[MAP_SIZE];

	if (*szMap && szMap)
	{
		hCmFile=fopen(szMap, "r");
		if (!hCmFile)
		{

			fread((void*) chBufCm, 1, MAP_SIZE, hCmFile );
			for (int i=0; i<ulLen && ucTextBuf<ucTextBufEnd; i++)
			{
				*ucTextBuf=chBufCm[ucText[i]];
				ucTextBuf++;
			}
		}
		else
		{
				if (szErr) sprintf(szErr, "can't open codemap file - %s !\n", szMap);
				return FALSE;
		}
	}
	else	return	FALSE;

	if (hCmFile)	fclose(hCmFile);

	return TRUE;
}
//---------------------------------------------------------------------------
unsigned long ConvertWithConstMap(char* szIn, char* szOut, unsigned char* ucText, unsigned long ulLen, unsigned char* ucTextBuf, unsigned long ulLenBuf, char* szErr)
{
	unsigned long i;
	unsigned char *pBufCm;//, *pBufCm2;

	if (!szIn || !szOut || !ulLen || !ucText || !ucTextBuf)
	{
		if (szErr)	sprintf(szErr, "wrong parametrs in function ConvertWithConstMap() !\n");
		return 0;
	}

	if (!stricmp(szIn, szOut) || !stricmp(szIn, "") || !stricmp(szOut, "") || !stricmp(szIn, CC_NONE) || !stricmp(szOut, CC_NONE))
	{
		if (ucText==ucTextBuf && ulLen==ulLenBuf)		return ulLen;
		else
		{
			for (i=0; i<ulLen && i<ulLenBuf; i++)
			{
				ucTextBuf[i]=ucText[i];
			}
		return	i;
		}
	}

	if (!IsSupportMap(szIn))
	{
		if (szErr)	sprintf(szErr, "%s - this codemap not support !\n", szIn);
		return 0;
	}

	if (!IsSupportMap(szOut))
	{
		if (szErr) sprintf(szErr, "%s - this codemap not support !\n", szOut);
		return 0;
	}

	if (!stricmp(szIn, CC_KOI8R))		pBufCm=aKOI8Rto1251_c;
	else if (!stricmp(szIn, CC_866))		pBufCm=a866to1251_c;
	else if (!stricmp(szIn, CC_DM))		pBufCm=aDMto1251_c;
	else if (!stricmp(szIn, CC_GOST))	pBufCm=aGOSTto1251_c;
	else if (!stricmp(szIn, CC_MAC))	pBufCm=aMACto1251_c;

	/*if (!stricmp(szIn, CC_UNICODE) && !stricmp(szOut, CC_UTF7))
	{
		UCS2* pBuf1=(UCS2*)ucText;
		char *pcBuf2=(char*) ucTextBuf;
		ConvertUCS2toUTF7(&pBuf1, (UCS2*) (pBuf1+ulLen), &pcBuf2, (char*) (pcBuf2+ulLenBuf), TRUE, TRUE);
		return (pcBuf2-((char*)ucTextBuf));

	}
	if (!stricmp(szIn, CC_UTF7) && !stricmp(szOut, CC_UNICODE))
	{
		UCS2* pBuf1=(UCS2*)ucText;
		char *pcBuf2=(char*) ucTextBuf;
		ConvertUTF7toUCS2(&pcBuf2, (char*) (pcBuf2+ulLenBuf), &pBuf1, (UCS2*) (pBuf1+ulLen));
		return (pBuf1-(UCS2*)ucText);

	}*/

	if (!stricmp(szIn, CC_1251) && !stricmp(szOut, CC_UNICODE))
	{
		return C1251toUNICODE(ucText, ulLen, ucTextBuf, ulLenBuf);

	}


	if (!stricmp(szIn, CC_1251) && !stricmp(szOut, CC_TRANSLIT))
	{
		return C1251toTranslit(ucText, ulLen, ucTextBuf, ulLenBuf, CN_TRANSLIT);

	}
	if (!stricmp(szIn, CC_1251) && !stricmp(szOut, CC_TRANSLIT_USA))
	{
		return C1251toTranslit(ucText, ulLen, ucTextBuf, ulLenBuf, CN_TRANSLIT_USA);

	}
	if (!stricmp(szIn, CC_1251) && !stricmp(szOut, CC_TRANSLIT_BGN))
	{
		return C1251toTranslit(ucText, ulLen, ucTextBuf, ulLenBuf, CN_TRANSLIT_BGN);

	}

	if (!stricmp(szOut, CC_1251) && !stricmp(szIn, CC_UNICODE))
	{
		return UNICODEtoC1251(ucText, ulLen, ucTextBuf, ulLenBuf);
	}

	if (!stricmp(szIn, CC_1251) || !stricmp(szIn, CC_UNICODE))
	{
		if (!stricmp(szOut, CC_KOI8R))		pBufCm=a1251toKOI8R_c;
		else if (!stricmp(szOut, CC_866))		pBufCm=a1251to866_c;
		else if (!stricmp(szOut, CC_DM))		pBufCm=a1251toDM_c;
		else if (!stricmp(szOut, CC_GOST))	pBufCm=a1251toGOST_c;
		else if (!stricmp(szOut, CC_MAC))	pBufCm=a1251toMAC_c;

	}

/*	if (!stricmp(szOut, CC_KOI8R))		pBufCm2=a1251toKOI8R_c;
	else if (!stricmp(szOut, CC_866))	pBufCm2=a1251to866_c;
	else if (!stricmp(szOut, CC_DM))		pBufCm2=a1251toDM_c;
	else if (!stricmp(szOut, CC_GOST))	pBufCm2=a1251toGOST_c;
	else if (!stricmp(szOut, CC_MAC))	pBufCm2=a1251toMAC_c;*/

	if (!stricmp(szIn, CC_TRANSLIT) || !stricmp(szIn, CC_TRANSLIT_USA) || !stricmp(szIn, CC_TRANSLIT_BGN))
	{
		if (szErr) sprintf(szErr, "can't convert text from %s to %s codemap !", szIn, szOut);
		return 0;
	}

	if (!stricmp(szOut, CC_1251) || !stricmp(szIn, CC_1251) || !stricmp(szOut, CC_UNICODE) || !stricmp(szIn, CC_UNICODE) || !stricmp(szOut, CC_TRANSLIT) || !stricmp(szOut, CC_TRANSLIT_USA) || !stricmp(szOut, CC_TRANSLIT_BGN))
	{
		for (i=0; i<ulLen && i<ulLenBuf; i++)
		{
			ucTextBuf[i]=pBufCm[ucText[i]];
		}

		if (!stricmp(szOut, CC_UNICODE))
		{
			return C1251toUNICODE(ucTextBuf, i, ucTextBuf, ulLenBuf);
		}
		else if (!stricmp(szIn, CC_UNICODE))
		{
			return UNICODEtoC1251(ucTextBuf, i, ucTextBuf, ulLenBuf);
		}
		else if (!stricmp(szOut, CC_TRANSLIT))
		{
			return C1251toTranslit(ucTextBuf, i, ucTextBuf, ulLenBuf, CN_TRANSLIT);
		}
		else if (!stricmp(szOut, CC_TRANSLIT_USA))
		{
			return C1251toTranslit(ucTextBuf, i, ucTextBuf, ulLenBuf, CN_TRANSLIT_USA);
		}
		else if (!stricmp(szOut, CC_TRANSLIT_BGN))
		{
			return C1251toTranslit(ucTextBuf, i, ucTextBuf, ulLenBuf, CN_TRANSLIT_BGN);
		}



	}
	else
	{
		if (szErr) sprintf(szErr, "can't convert text from %s to %s codemap !", szIn, szOut);
		return 0;
/*		for (i=0; i<ulLen; i++)
		{
			ucTextBuf[i]=pBufCm[ucText[i]];
			ucTextBuf[i]=pBufCm2[ucText[i]];
		}*/
	}

return ulLen;
}
//---------------------------------------------------------------------------
BOOL DefineMap(unsigned char* ucText, unsigned long ulLen, char* szMap)
{
	unsigned int map[MAPS_AUTO_N];

	unsigned long i=0;
	unsigned char a;
	unsigned int k=0;
	unsigned int utf=0;
	int max=-1;
	BOOL fOnlyAscii=TRUE, fSS=FALSE;

	memset(map, 0, sizeof(map));

	if (!ucText || !ulLen || !szMap)	return FALSE;

	while (i<ulLen)
	{
		a=*(ucText+i);

		if (!a || a==1 || a==4)
		{
			if (ulLen%sizeof(unsigned short)+1)	utf++;
		}

		if (a>=0 && a!=9 && a!=13 && a!=10 && a<=31)
		{
			fSS=TRUE;
		}
		else if (a>=MAX_LATCHAR)
		{
			fOnlyAscii=FALSE;
		}

		if (a>=MAX_LATCHAR)
		{
			k++;
			if (a==238 || a==224 || a==229) map[0]++;
			if (a==174 || a==160 || a==165) map[1]++;
			if (a==237 || a==185 || a==228) map[2]++;
			if (a==207 || a==193 || a==197) map[3]++;
			if (a==222 || a==208 || a==213) map[4]++;
		//	if (a==222 || a==208 || a==213) map[5]++; Problems with MAC codemap !!!
		}
		i++;
		if (k>MAX_RUS_SYMBOLS)	break;
	}

	// This is Unicode !!!
	if (utf>2 && !(ulLen%sizeof(unsigned short)) && utf>(i/sizeof(unsigned short)-2))
	{
		strcpy(szMap, CC_UNICODE);
		return TRUE;
	}

	if (fOnlyAscii)
	{
		strcpy(szMap, CC_1251);
		return TRUE;
	}

	if (fSS)
	{
		*szMap=0;
		return FALSE;
	}

	unsigned int buf=0;
	for (i=0; i<MAPS_AUTO_N; i++)
	{
		if (map[i]>buf)
		{
			buf=map[i];
			max=(int) i;
		}
	}

	switch(max)
	{
		case 0:
			strcpy(szMap, CC_1251);
		break;
		case 1:
			strcpy(szMap, CC_866);
		break;
		case 2:
			strcpy(szMap, CC_DM);
		break;
		case 3:
			strcpy(szMap, CC_KOI8R);
		break;
		case 4:
			strcpy(szMap, CC_GOST);
		break;
		case 5:
			strcpy(szMap, CC_MAC);
		break;
		default:
			*szMap=0;
			return FALSE;
	}

	return TRUE;
}
//---------------------------------------------------------------------------
BOOL IsSupportMap(char* szMap)
{
	int i;
	if (!szMap || !*szMap) return FALSE;

	for (i=0; i<MAPS_N; i++)
	{
		if (!stricmp(szMap, pMaps[i]))
		{
			return TRUE;
		}
	}

	return FALSE;
}
//---------------------------------------------------------------------------
unsigned short TurnShort(unsigned short usShort)
{
	unsigned short usBuf=usShort;
	unsigned char* pucBuf;
	unsigned char ucBuf;

	pucBuf=(unsigned char*) &usBuf;
	ucBuf=*pucBuf;
	*pucBuf=*(pucBuf+1);
	*(pucBuf+1)=ucBuf;

return usBuf;
}
//---------------------------------------------------------------------------
unsigned long C1251toUNICODE(unsigned char* ucText, unsigned long ulLen, unsigned char* ucTextBuf, unsigned long ulLenBuf)
{
	unsigned long i,n, nmax;
	static unsigned short utfbuf[MAX_USHORT];

	unsigned short* utf=(unsigned short*) ucTextBuf;
	nmax=ulLenBuf-ulLenBuf%(sizeof(unsigned short));

	static BOOL fInit=FALSE;
	ucTextBuf[0]=0xFF;
	ucTextBuf[1]=0xFE;
	utf++;

	if (ulLenBuf<(sizeof(unsigned short)*2) || !ulLen) return 0;
	if (!ucText || !ucTextBuf) return 0;

	if (!fInit)
	{
		memset(utfbuf, 0, sizeof(utfbuf));

		for (i=0; i<MAX_LATCHAR; i++)
		{
			utfbuf[i]=TurnShort((unsigned short) i);
		}

		for (i=MAX_LATCHAR; i<MAX_UCHAR; i++)
		{
			for (n=0; n<TABLE_1251_UNICODE_SIZE; n++)
			{
				if (Table_1251_UNICODE[n][0]==i)
				{
					utfbuf[i]=TurnShort(Table_1251_UNICODE[n][1]);
					break;
				}
			}
		}
		fInit=TRUE;
	}

	n=0;
	for (i=0; i<ulLen && n<nmax; i++)
	{
		utf[n]=utfbuf[ucText[i]];
		n++;
	}

	return (sizeof(unsigned short)*(n+1));
}
//---------------------------------------------------------------------------
unsigned long UNICODEtoC1251(unsigned char* ucText, unsigned long ulLen, unsigned char* ucTextBuf, unsigned long ulLenBuf)
{
	unsigned long i, n;
	unsigned short* utf=(unsigned short*) ucText;

	static unsigned char buf1251[MAX_USHORT];
	static BOOL fInit=FALSE;

	if (ucText[0]==0xFF && ucText[1]==0xFE)
	{
		utf++;
		ucText+=sizeof(unsigned short);
		ulLen-=sizeof(unsigned short);
	}
	if (!fInit)
	{
		memset(buf1251, 0, sizeof(buf1251));

		for (i=0; i<MAX_LATCHAR; i++)
		{
			buf1251[i]=(unsigned char) i;
		}

		for (i=0; i<TABLE_1251_UNICODE_SIZE; i++)
		{
			buf1251[Table_1251_UNICODE[i][1]]=(unsigned char) Table_1251_UNICODE[i][0];
		}
		fInit=TRUE;
	}

	n=0;
	for (i=0; i<ulLen/2 && n<ulLenBuf; i++)
	{
		ucTextBuf[n]=buf1251[TurnShort(utf[i])];
		n++;
	}

	return n;
}
//---------------------------------------------------------------------------
unsigned long CalcBufMap(char* szIn, char* szOut, unsigned char* ucText, unsigned ulLen)
{
	int i, iIn=-1, iOut=-1;

	for (i=0; i<MAPS_N; i++)
	{
		if (!stricmp(pMaps[i], szIn))
		{
			iIn=i;
		}
		if (!stricmp(pMaps[i], szOut))
		{
			iOut=i;
		}
	}
	if (iOut==CN_NONE || iIn==CN_NONE)	return ulLen;

	if (iOut==-1 || iIn==-1) return 0;

	//printf("iIn=%d, iOut=%d\n", iIn, iOut);

	if (iIn==iOut)	return ulLen;

	if (!stricmp(CC_TRANSLIT, szIn) || !stricmp(CC_TRANSLIT_USA, szIn) || !stricmp(CC_TRANSLIT_BGN, szIn))
	{

		return ((ulLen+2)*SizeMap[iOut]);

	}

	return ((ulLen+2)*SizeMap[iOut]/SizeMap[iIn]);
}
//---------------------------------------------------------------------------
unsigned long C1251toTranslit(unsigned char* ucText, unsigned long ulLen, unsigned char* ucTextBuf, unsigned long ulLenBuf, int iMode)
{
	unsigned long i,n, nmax;
	char* transbuf[MAX_UCHAR], *cBuf=NULL;
      char ** ppcTranslitBuf;

	static BOOL fInit=FALSE;

	if (!ulLen || !ulLenBuf) return 0;

	if (!ucText || !ucTextBuf) return 0;

	ppcTranslitBuf=Translit_Simple;

	if (iMode==CN_TRANSLIT_USA)	ppcTranslitBuf=Translit_USA;
	if (iMode==CN_TRANSLIT_BGN)	ppcTranslitBuf=Translit_BGN;

	if (!fInit)
	{
		memset(transbuf, 0, sizeof(transbuf));

		n=sizeof(Translit_Main)/(sizeof(unsigned char));

		for (i=0; i<n; i++)
		{
			transbuf[Translit_Main[i]]=ppcTranslitBuf[i];
		}
		fInit=TRUE;
	}

	n=0;
	for (i=0; i<ulLen && n<ulLenBuf; i++)
	{
		cBuf=transbuf[ucText[i]];

		if (!cBuf)
		{
			ucTextBuf[n]=(unsigned char) ucText[i];
			n++;
		}
		else
		{
			strcpy((char *) &ucTextBuf[n], cBuf);
			n+=strlen(cBuf);
		}

	}
	//printf("ulLen=%d, n=%d, ulLenBuf=%d\n", ulLen, n, ulLenBuf);

	return n;
}